#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2023 Luke Zaruba
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""overlay.py: Overlay analyses for use on clusters of footprints from periods ("snapshots") of time to assess the temporal stability of clusters."""

import math

from geopandas import GeoDataFrame, sjoin
from pandas import merge
from shapely.ops import polygonize


def weighted_overlay(footprints: GeoDataFrame) -> GeoDataFrame:
    """
    Overlay approach that takes time into account and weights more recent clusters at a higher level than less recent clusters.

    Computed as the period index divided by the sum of all period indexes.

    Args:
        footprints (GeoDataFrame): Input polygon fetaures from TemporalADBSCAN.

    Returns:
        GeoDataFrame: Output polygons features.
    """
    # Reset Index
    footprints.reset_index(inplace=True)

    # Find Unique Periods & Sort
    unique_periods = [str(i) for i in footprints.PERIOD.unique()]
    unique_periods.sort()

    # Create Dictionary of Periods & Indexes
    period_indexes = {}

    for index, value in enumerate(unique_periods):
        period_indexes[value] = index + 1

    # Calculate Sum of Indexes
    index_sum = sum(list(period_indexes.values()))

    # Compute Weights in DF
    footprints["weight"] = 0.0

    for index, row in footprints.iterrows():
        footprints.at[index, "weight"] = period_indexes[row["PERIOD"]] / index_sum

    # Find Bounds of Footprints & Convert to New GDF
    bounds = footprints.geometry.exterior.unary_union
    polygons = GeoDataFrame(geometry=list(polygonize(bounds)), crs=footprints.crs)
    polygons["OID"] = range(len(polygons))

    # Get Centroids
    polygons_centroid = polygons.copy()
    polygons_centroid["geometry"] = polygons.centroid

    # Spatial Join & Counting
    counts = sjoin(polygons_centroid, footprints)
    counts = counts.groupby(["OID"])["weight"].sum().rename("WEIGHT").reset_index()

    # Return GeoDataFrame with Counts
    return merge(polygons, counts)


def simple_overlay(footprints: GeoDataFrame) -> GeoDataFrame:
    """
    Overlay approach that finds overlapping polygons and calculates the feature count for each intersecting area.

    Args:
        footprints (GeoDataFrame): Input polygon fetaures from TemporalADBSCAN.

    Returns:
        GeoDataFrame: Output polygons features.
    """
    # Find Bounds of Footprints & Convert to New GDF
    bounds = footprints.geometry.exterior.unary_union
    polygons = GeoDataFrame(geometry=list(polygonize(bounds)), crs=footprints.crs)
    polygons["OID"] = range(len(polygons))

    # Get Centroids
    polygons_centroid = polygons.copy()
    polygons_centroid["geometry"] = polygons.centroid

    # Spatial Join & Counting
    counts = sjoin(polygons_centroid, footprints)
    counts = (
        counts.groupby(["OID"])["index_right"].count().rename("COUNT").reset_index()
    )

    # Return GeoDataFrame with Counts
    return merge(polygons, counts)


def _calculate_cluster_time_indexes(footprints: GeoDataFrame) -> GeoDataFrame:
    """
    Function to create list of time series indexes where cluster was defined as cluster and store back into GDF.

    Args:
        footprints (GeoDataFrame): Input polygon features.

    Returns:
        GeoDataFrame: Output polygon features with time index lists.
    """
    # Reset Index
    footprints = footprints.reset_index(drop=True)

    # Find Unique Periods & Sort
    unique_periods = [str(i) for i in footprints["PERIOD"].unique()]
    unique_periods.sort()

    # Create Dictionary of Periods & Indexes
    period_indexes = {}

    for index, value in enumerate(unique_periods):
        period_indexes[value] = index + 1

    # Compute Weights in DF
    footprints["index"] = 0.0

    for index, row in footprints.iterrows():
        footprints.at[index, "index"] = period_indexes[row["PERIOD"]]

    # Find Bounds of Footprints & Convert to New GDF
    bounds = footprints.geometry.exterior.unary_union
    polygons = GeoDataFrame(geometry=list(polygonize(bounds)), crs=footprints.crs)
    polygons["OID"] = range(len(polygons))

    # Get Centroids
    polygons_centroid = polygons.copy()
    polygons_centroid["geometry"] = polygons.centroid

    # Spatial Join
    counts = sjoin(polygons_centroid, footprints)

    # Find Indexes for Each Cluster & Store in Dict
    cluster_index_dict = {}

    for oid in counts.OID.unique():
        oid_indexes = list(counts.loc[counts["OID"] == oid]["index"])

        oid_indexes = [int(i) for i in oid_indexes]

        oid_indexes.sort()

        cluster_index_dict[oid] = oid_indexes

    # Create New Column with Empty List for Each Row
    polygons["index_list"] = [[] for _ in range(len(polygons))]

    # Update Column to List of Indexes where included as Cluster
    for index, row in polygons.iterrows():
        try:
            polygons.at[index, "index_list"] = cluster_index_dict[row["OID"]]
        except KeyError:
            print(f"KeyError: {row['OID']} not found")
            polygons.at[index, "index_list"] = []

    # Return
    return polygons, list(period_indexes.values())


def stcec(footprints: GeoDataFrame, significance_percent: float = 0.9) -> GeoDataFrame:
    """
    Function to classify index list of a given cluster area into an STCEC classification.

    Args:
        footprints (GeoDataFrame): Input polygon features.
        significance_percent (float, optional): Percentage used to determine temporal significance of long lasting clusters. Defaults to 0.9.

    Returns:
        GeoDataFrame: Output polygon features with STCEC classes.
    """
    # Create List of Indexes
    footprints_new, indexes = _calculate_cluster_time_indexes(footprints)

    # Index Stats
    indexes.sort()

    total_indexes = len(indexes)
    significance_length = math.ceil(total_indexes * significance_percent)

    most_recent_index = indexes[-1:][0]
    second_most_recent_index = indexes[-2:-1][0]

    # Classification
    footprints_new["STCEC"] = ""

    for index, row in footprints_new.iterrows():
        # Get List of Indexes for each row
        index_list = row["index_list"]

        index_list.sort()

        # Checks for Determining STCEC Class
        if len(index_list) >= significance_length:
            if most_recent_index in index_list:
                value = "PERSISTENT"
            else:
                value = "HISTORICAL"
        else:
            if len(index_list) == 1:
                if most_recent_index in index_list:
                    value = "NEW"
                else:
                    value = "SPORADIC"
            else:
                if (
                    most_recent_index in index_list
                    and second_most_recent_index in index_list
                ):
                    value = "INTENSIFYING"
                else:
                    for i in range(len(index_list)):
                        try:
                            consecutive = index_list[i] + 1 == index_list[i + 1]
                            if consecutive:
                                value = "DIMINISHING"
                                break
                        except IndexError:
                            value = "OCCASIONAL"

        # Set STCEC Value
        footprints_new.at[index, "STCEC"] = value

    # Return new GDF
    return footprints_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2023 Luke Zaruba
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""test_clustering.py: Testing for stco/clustering.py - should be run from ./tests using `python -m unittest -v test_clustering.py`"""

import sys
import unittest
from pathlib import Path

from geopandas import GeoDataFrame, points_from_xy
from numpy.random import seed
from pandas import DataFrame, concat, to_datetime
from shapely import Polygon
from sklearn.datasets import make_blobs

sys.path.append(str(Path(__file__).parent.parent.parent))

from stco.clustering import TemporalADBSCAN  # noqa: E402


def _create_dataset():
    # Set of Blobs for Time Step 1
    X_1, _ = make_blobs(
        n_samples=100, centers=[[1, 1]], cluster_std=0.4, random_state=0
    )
    df_1 = DataFrame(
        {"X": [i[0] for i in X_1], "Y": [i[1] for i in X_1], "time": "01/01/2021"}
    )
    df_1["time"] = to_datetime(df_1["time"])
    gdf_1 = GeoDataFrame(df_1, geometry=points_from_xy(df_1["X"], df_1["Y"]))

    # Set of Blobs for Time Step 2
    X_2, _ = make_blobs(
        n_samples=100, centers=[[0.5, 0.75]], cluster_std=0.4, random_state=1
    )
    df_2 = DataFrame(
        {"X": [i[0] for i in X_2], "Y": [i[1] for i in X_2], "time": "01/01/2022"}
    )
    df_2["time"] = to_datetime(df_2["time"])
    gdf_2 = GeoDataFrame(df_2, geometry=points_from_xy(df_2["X"], df_2["Y"]))

    # Set of Blobs for Time Step 3
    X_3, _ = make_blobs(
        n_samples=100, centers=[[0.75, 1.5]], cluster_std=0.4, random_state=2
    )
    df_3 = DataFrame(
        {"X": [i[0] for i in X_3], "Y": [i[1] for i in X_3], "time": "01/01/2023"}
    )
    df_3["time"] = to_datetime(df_3["time"])
    gdf_3 = GeoDataFrame(df_3, geometry=points_from_xy(df_3["X"], df_3["Y"]))

    # Concat
    gdf = concat([gdf_1, gdf_2, gdf_3])

    return gdf


class TestTemporalADBSCAN(unittest.TestCase):
    def setUp(self):
        # Data Set Up
        self.gdf = _create_dataset()
        seed(10)
        self.tabds = TemporalADBSCAN(
            self.gdf, "Y", 0.5, 0.05, pct_exact=0.5, reps=50, keep_solus=True
        )

        # Solutions
        self.wkt_solu = "POLYGON ((0.7425526388668438 0.1106387391102293, 0.6448857009479549 0.2076814127104292, 0.5999138610441741 0.3820915612889553, 0.5805788139731629 0.4319928251284099, 0.4736370357953915 0.8153661580741164, 0.3544408609768194 0.9149038879144125, 0.3479206612135821 1.1851129022103097, 0.4918060006057066 1.3877586832632045, 0.3174919237499949 1.7803101580927159, 0.5307506379543361 1.777448474259717, 0.655509725978119 1.7640259812396135, 0.800787019723078 1.7718128215267943, 1.0042000082883282 1.7143481975623343, 1.3914951936422957 1.8963572796805832, 1.5953008775182398 1.7583556704122332, 1.769176810592154 1.5922059165737696, 1.8652943797122759 1.534611179774557, 1.953257909945577 1.3777917947961655, 1.7056209383870655 1.1600628833468893, 1.5976316292630424 0.9179366944936797, 1.747023196059987 0.6090888480494356, 1.907901849595043 0.418253730160494, 1.7532602788225018 0.4608963755430214, 1.4557602738173203 0.5060696718585389, 1.250092580410875 0.3591769377573009, 1.186264975892184 0.3855025254891105, 0.8584024354986064 0.4500194826327925, 0.7425526388668438 0.1106387391102293))"
        self.len_solu = 3

    def test_fit_all(self):
        # Run Fit
        footprints = self.tabds.fit_all()

        # Assert
        self.assertEqual(footprints.iloc[0]["geometry"].wkt, self.wkt_solu)
        self.assertEqual(len(footprints), self.len_solu)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2023 Luke Zaruba
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""test_overlay.py: Testing for stco/overlay.py - should be run from ./tests using `python -m unittest -v test_overlay.py`"""

import sys
import unittest
from pathlib import Path

from geopandas import GeoDataFrame
from shapely import Polygon

sys.path.append(str(Path(__file__).parent.parent.parent))

from stco.overlay import simple_overlay, stcec, weighted_overlay  # noqa: E402


class TestSimpleOverlay(unittest.TestCase):
    def setUp(self):
        # Data
        polygon_1 = Polygon(
            ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0))
        )
        polygon_2 = Polygon(
            ((0.5, 0.0), (0.5, 1.0), (1.5, 1.0), (1.5, 0.0), (0.5, 0.0))
        )
        polygon_3 = Polygon(
            ((0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5), (0.5, 0.5))
        )

        data = {
            "PERIOD": ["2021", "2022", "2023"],
            "geometry": [polygon_1, polygon_2, polygon_3],
        }

        self.gdf = GeoDataFrame(data)
        self.wkt_solu = "POLYGON ((0 0, 0 1, 0.5 1, 0.5 0.5, 0.5 0, 0 0))"
        self.count_solu = 1

    def test_simple_overlay(self):
        # Run Simple Overlay
        simple_overlay_output = simple_overlay(self.gdf)

        # Assert
        self.assertEqual(simple_overlay_output.iloc[0]["geometry"].wkt, self.wkt_solu)
        self.assertEqual(simple_overlay_output.iloc[0]["COUNT"], self.count_solu)


class TestSTCEC(unittest.TestCase):
    def setUp(self):
        # Data
        polygon_1 = Polygon(
            ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0))
        )
        polygon_2 = Polygon(
            ((0.5, 0.0), (0.5, 1.0), (1.5, 1.0), (1.5, 0.0), (0.5, 0.0))
        )
        polygon_3 = Polygon(
            ((0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5), (0.5, 0.5))
        )

        data = {
            "PERIOD": ["2021", "2022", "2023"],
            "geometry": [polygon_1, polygon_2, polygon_3],
        }

        self.gdf = GeoDataFrame(data)
        self.wkt_solu = "POLYGON ((0 0, 0 1, 0.5 1, 0.5 0.5, 0.5 0, 0 0))"
        self.stcec_solu = "SPORADIC"

    def test_stcec(self):
        # Run STCEC
        stcec_output = stcec(self.gdf)

        # Assert
        self.assertEqual(stcec_output.iloc[0]["geometry"].wkt, self.wkt_solu)
        self.assertEqual(stcec_output.iloc[0]["STCEC"], self.stcec_solu)


class TestWeightedOverlay(unittest.TestCase):
    def setUp(self):
        # Data
        polygon_1 = Polygon(
            ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0))
        )
        polygon_2 = Polygon(
            ((0.5, 0.0), (0.5, 1.0), (1.5, 1.0), (1.5, 0.0), (0.5, 0.0))
        )
        polygon_3 = Polygon(
            ((0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5), (0.5, 0.5))
        )

        data = {
            "PERIOD": ["2021", "2022", "2023"],
            "geometry": [polygon_1, polygon_2, polygon_3],
        }

        self.gdf = GeoDataFrame(data)
        self.wkt_solu = "POLYGON ((0 0, 0 1, 0.5 1, 0.5 0.5, 0.5 0, 0 0))"
        self.weight_solu = 0.167

    def test_simple_overlay(self):
        # Run Weighted Overlay
        weighted_overlay_output = weighted_overlay(self.gdf)

        # Assert
        self.assertEqual(weighted_overlay_output.iloc[0]["geometry"].wkt, self.wkt_solu)
        self.assertEqual(
            round(weighted_overlay_output.iloc[0]["WEIGHT"], 3), self.weight_solu
        )


suite = unittest.TestSuite()
for i in [TestSimpleOverlay, TestSTCEC, TestWeightedOverlay]:
    a = unittest.TestLoader().loadTestsFromTestCase(i)
    suite.addTest(a)

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2023 Luke Zaruba
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""clustering.py: Implementation of A-DBSCAN on periods ("snapshots") of a time series for use in subsequent overlay analyses."""

from math import pi, sqrt
from typing import Union

from esda.adbscan import ADBSCAN, get_cluster_boundary
from geopandas import GeoDataFrame
from pandas import DataFrame, concat, to_datetime
from tqdm import tqdm


class TemporalADBSCAN:
    def __init__(
        self,
        data: Union[DataFrame, GeoDataFrame],
        period: str,
        eps: float,
        min_sample_pct: float,
        time_field: str = "time",
        x_field: str = "X",
        y_field: str = "Y",
        **kwargs,
    ) -> None:
        """
        Instantiates TemporalADBSCAN object to run snapshot-based A-DBSCAN.

        Args:
            data (Union[DataFrame, GeoDataFrame]): Input point features.
            period (str): Pandas/NumPy Period alias.
            eps (float): Eps parameter for ADBSCAN, in the unit of the input dataset's coordinate system.
            min_sample_pct (float): Percentage of the total points within each period to use as the minimum samples parameter in ADBSCAN.
            time_field (str, optional): Name of time field in data. Defaults to "time".
            x_field (str, optional): Name of X coordinate field in data - not needed for GeoDataFrame inputs. Defaults to "X".
            y_field (str, optional): Name of Y coordinate field in data - not needed for GeoDataFrame inputs. Defaults to "Y".
        """
        if type(data) is GeoDataFrame:
            data = self._create_xy_series(data)
        self.data = self._check_data_types(data, time_field, x_field, y_field)
        self.period = self._check_period(period)
        self.eps = eps
        self.min_sample_pct = min_sample_pct
        self.time_field = time_field
        self.x_field = x_field
        self.y_field = y_field

        # Handling other kwargs
        self._adbscan_algorithm = kwargs.get("algorithm", "auto")
        self._adbscan_n_jobs = kwargs.get("n_jobs", 1)
        self._adbscan_pct_exact = kwargs.get("pct_exact", 0.1)
        self._adbscan_reps = kwargs.get("reps", 100)
        self._adbscan_keep_solus = kwargs.get("keep_solus", False)
        self._adbscan_pct_thr = kwargs.get("pct_thr", 0.9)

    @staticmethod
    def _check_period(period_value: str) -> str:
        """
        Method to check the validity of the period value passed in.

        Args:
            period_value (str): Period value to create "snapshots" with.

        Raises:
            ValueError: Raised when invalid period value is passed in.

        Returns:
            str: Period value, if valid.
        """
        valid_periods = [
            "Y",  # Annual
            "Q",  # Quarterly
            "M",  # Monthly
            "W",  # Weekly
            "D",  # Daily
            "H",  # Hourly
            "T",  # Minutely
        ]

        if period_value not in valid_periods:
            raise ValueError(
                f"Period must be one of following strings: {valid_periods}"
            )

        else:
            return period_value

    @staticmethod
    def _check_data_types(
        data: Union[DataFrame, GeoDataFrame],
        time_field: str,
        x_field: str,
        y_field: str,
    ) -> Union[DataFrame, GeoDataFrame]:
        """
        Method to check data types and attempt to fix any issues.

        Args:
            data (Union[DataFrame, GeoDataFrame]): Input point features.
            time_field (str): Name of time field in data.
            x_field (str): Name of X coordinate field in data.
            y_field (str): Name of Y coordinate field in data.

        Raises:
            TypeError: Raised when type of time field passed in in invalid.
            TypeError: Raised when type of X coordinate field passed in in invalid.
            TypeError: Raised when type of Y coordinate field passed in in invalid.

        Returns:
            Union[DataFrame, GeoDataFrame]: Output data, potentially updated with new types.
        """
        # Check Time
        if data[time_field].dtype.name != "datetime64[ns]":
            try:
                data[time_field] = to_datetime(data[time_field])
            except TypeError:
                print(
                    f"Series for Time field, {time_field}, must be of type 'datetime64[ns]'"
                )

        # Check X
        if data[x_field].dtype.name != "float64":
            try:
                data[x_field] = data[x_field].astype("float64")
            except TypeError:
                print(f"Series for X field, {x_field}, must be of type 'float64'")

        # Check Y
        if data[y_field].dtype.name != "float64":
            try:
                data[y_field] = data[y_field].astype("float64")
            except TypeError:
                print(f"Series for Y field, {y_field}, must be of type 'float64'")

        # Return updated version of data -- in case of casting
        return data

    @staticmethod
    def _create_xy_series(data: GeoDataFrame) -> GeoDataFrame:
        """
        Method to convert GeoDataFrame Point coordinates to X and Y series in GeoDataFrame.

        Args:
            data (GeoDataFrame): Input point features.

        Returns:
            GeoDataFrame: Output point features.
        """
        if "X" not in data.columns:
            data["X"] = data.geometry.x
        if "Y" not in data.columns:
            data["Y"] = data.geometry.y

        return data

    def _create_period(self) -> None:
        """
        Creates a new Series in the DataFrame containing the period of the record.
        """
        self.data["PERIOD"] = (
            self.data[self.time_field].dt.to_period(self.period).astype(str)
        )

    def fit_all(self) -> GeoDataFrame:
        """
        Method to fit model to data from all periods/snapshots of time, create cluster boundaries, and concatenate footprints results into single GeoDataFrame.

        Returns:
            GeoDataFrame: Output cluster footprints.
        """
        # Import Numba for Boundary Construction
        import numba

        # Create Periods
        self._create_period()

        # Create List of Empty Output GDFs
        footprint_df_list = []

        # Loop through Timesteps
        for step in tqdm(self.data.PERIOD.unique()):
            # Filter data
            filtered_data = self.data.loc[(self.data["PERIOD"] == step)].copy()

            # Fit Model
            abdscan = ADBSCAN(
                self.eps,
                filtered_data.shape[0] * self.min_sample_pct,
                self._adbscan_algorithm,
                self._adbscan_n_jobs,
                self._adbscan_pct_exact,
                self._adbscan_reps,
                self._adbscan_keep_solus,
                self._adbscan_pct_thr,
            )

            abdscan.fit(filtered_data)

            # Get Footprints
            footprints = get_cluster_boundary(abdscan.votes["lbls"], filtered_data)

            # Create Dictionary of Timestep Label & Footprints
            data = {"PERIOD": step, "geometry": footprints}

            # Convert to GDF & Append
            footprint_df_list.append(GeoDataFrame(data))

        # Concatenate GDFs
        return concat(footprint_df_list)
